#include "scripts/BPR/BPR_Shared.cpp"

#include "scripts/BPR/BPR_MapData.cpp"
#include "scripts/BPR/BPR_Tasks.cpp"
#include "scripts/BPR/BPR_WorldObject.cpp"
#include "scripts/BPR/BPR_ConsoleCommands.cpp"

namespace BPR
{
    const array<int> g_niBossMapIDs = { 0, 3, 48, 49 };
    const array<int> g_niStartMapIDs = { 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 48, 49 };
    const array<kStr> g_nsStartMapPaths = {
        "levels/level04.map", "levels/level05.map", "levels/level06.map", "levels/level07.map", "levels/level08.map",
        "levels/level09.map", "levels/level10.map", "levels/level11.map", "levels/level12.map", "levels/level13.map",
        "levels/level14.map", "levels/level15.map", "levels/level16.map", "levels/level17.map", "levels/level18.map",
        "levels/level19.map", "levels/level20.map", "levels/level48.map", "levels/level49.map"
    };
    
    bool g_bStartedGame = false;
    bool g_bWarpingPlayer = false;
    bool g_bGiveStartingWeapons = false;
    bool g_bForceLevelPortalActivate = false;
    bool g_bSpawningActor = false;
	kPuppet@ g_pPlayerPuppet = null;
	kActor@ g_pPlayerActor = null;
	array<kActor@> g_npActors;
    int g_iSeedStartMap;
    int g_iSeedWarp;
    int g_iSeedHubPortals;
    int g_iBowShotType;
    kVec3 g_vPlayerStartPos;
    int g_iPlayerStartSector;
    float g_fPlayerStartYaw;
    int g_iShowStatsTime;
    
	//------------------------------------------------------------------------------------------------------------------------
	// Called from main(void)
	//------------------------------------------------------------------------------------------------------------------------
	void OnMain()
	{
        Sys.Print("--------------------------------------------------------------------------------------------------------------------------------");
        kStr sModTitle = "Randomizer v";
        Sys.Warning(sModTitle + VERSION);
        Sys.Warning("By: BehemothProgrammer");
        Sys.Print("--------------------------------------------------------------------------------------------------------------------------------");
        BPR::AddGameVariables();
        SetStartSeed(0);
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Called from newgame(void)
	//------------------------------------------------------------------------------------------------------------------------
	void OnNewGame()
	{
        BPR::AddGameVariables();
        GameVariables.SetValue(GV_KEY_VERSION, SAVE_VERSION);
        GameVariables.SetValue(GV_KEY_TICKS, "0");
		GameVariables.SetValue(GV_KEY_STARTED, "0");
		GameVariables.SetValue(GV_KEY_PWARP, "0");
		GameVariables.SetValue(GV_KEY_LOWEST_DIFFICULTY, "3");
        GameVariables.SetValue(GV_KEY_KNIFE_POINTER, "0");
        
        for(int i = 0; i < 100; ++i)
        {
            GameVariables.SetValue(GV_KEY_MAPDATA + i, "");
        }
	}
	//------------------------------------------------------------------------------------------------------------------------
    void AddGameVariables()
    {
        if(!GameVariables.HasKey(GV_KEY_VERSION)) GameVariables.Add(GV_KEY_VERSION, SAVE_VERSION);
        if(!GameVariables.HasKey(GV_KEY_START_SEED)) GameVariables.Add(GV_KEY_START_SEED, "0");
        if(!GameVariables.HasKey(GV_KEY_TICKS)) GameVariables.Add(GV_KEY_TICKS, "0");
        if(!GameVariables.HasKey(GV_KEY_STARTED)) GameVariables.Add(GV_KEY_STARTED, "0");
        if(!GameVariables.HasKey(GV_KEY_PWARP)) GameVariables.Add(GV_KEY_PWARP, "0");
        if(!GameVariables.HasKey(GV_KEY_LOWEST_DIFFICULTY)) GameVariables.Add(GV_KEY_LOWEST_DIFFICULTY, "0");
        if(!GameVariables.HasKey(GV_KEY_KNIFE_POINTER)) GameVariables.Add(GV_KEY_KNIFE_POINTER, "0");
        if(!GameVariables.HasKey(GV_KEY_OPT_PLAYER_LOC)) GameVariables.Add(GV_KEY_OPT_PLAYER_LOC, "1");
        if(!GameVariables.HasKey(GV_KEY_OPT_ALWAYS_KNIFE)) GameVariables.Add(GV_KEY_OPT_ALWAYS_KNIFE, "0");
        if(!GameVariables.HasKey(GV_KEY_OPT_ALWAYS_BOW)) GameVariables.Add(GV_KEY_OPT_ALWAYS_BOW, "0");
    }
	//------------------------------------------------------------------------------------------------------------------------
	// Called from TurokPlayer::constructor
	//------------------------------------------------------------------------------------------------------------------------
	void OnMapInit(kPuppet @pPlayerPuppet)
	{
        BPR::AddGameVariables();
        g_vPlayerStartPos = pPlayerPuppet.Origin();
        g_iPlayerStartSector = pPlayerPuppet.SectorIndex();
        g_fPlayerStartYaw = pPlayerPuppet.Yaw();
        GetSeedValues();
		GameVariables.GetBool(GV_KEY_STARTED, g_bStartedGame);
		GameVariables.GetBool(GV_KEY_PWARP, g_bWarpingPlayer);
        g_iShowStatsTime = 0;
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Called from TurokPlayer::OnSpawn
	//------------------------------------------------------------------------------------------------------------------------
	void OnMapSpawn()
	{
        
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Called from TurokPlayer::OnBeginLevel
	//------------------------------------------------------------------------------------------------------------------------
	void OnMapBegin()
	{
		@g_pPlayerPuppet = Player.Actor();
		@g_pPlayerActor = g_pPlayerPuppet.CastToActor();
        g_bGiveStartingWeapons = false;
        
        if(!g_bStartedGame && Game.GetCurrentMapID() != 5 && Game.GetCurrentMapID() != 42 &&
            Game.GetCurrentMapID() != 43 && Game.GetCurrentMapID() != 44 &&
            Game.GetCurrentMapID() != 45 && Game.GetCurrentMapID() != 46)
        {
            if(Game.GetCurrentMapID() == 4)
            {
                BPR::Tasks::Add(PlayerWarpCheck);
            }
            else
            {
                SetGameStarted();
                g_bGiveStartingWeapons = true;
            }
        }
        
        switch(Game.GetCurrentMapID())
		{
        case 42:
            GameVariables.SetValue(GV_KEY_KNIFE_POINTER, "1");
            break;
        }
        
        //warp player to random warp point in map
        if(g_bWarpingPlayer)
        {
            g_bWarpingPlayer = false;
            g_bGiveStartingWeapons = true;
            GameVariables.SetValue(GV_KEY_PWARP, "0");
            
            if(BPR::GameVar::GetBool(GV_KEY_OPT_PLAYER_LOC))
            {
                array<kActor@> npValidActors;
                npValidActors.insertLast(g_pPlayerActor);
                for(uint i = 0; i < g_npActors.length(); ++i)
                {
                    kActor@ pActor = g_npActors[i];
                    if(pActor.Type() != AT_WARPPOINT)
                    {
                        continue;
                    }
                    npValidActors.insertLast(pActor);
                }
                if(npValidActors.length() > 0)
                {
                    kActor@ pActor = npValidActors[BPR::RandMax(g_iSeedWarp, int(npValidActors.length()))];
                    if(pActor.Type() == AT_WARPPOINT)
                    {
                        g_pPlayerActor.SetSector(pActor.SectorIndex());
                        g_pPlayerActor.SetPosition(pActor.Origin());
                        g_pPlayerActor.Yaw() = pActor.Yaw() + Math::pi;
                    }
                    else //Is player Start Position
                    {
                        g_pPlayerActor.SetSector(g_iPlayerStartSector);
                        g_pPlayerActor.SetPosition(g_vPlayerStartPos);
                        g_pPlayerActor.Yaw() = g_fPlayerStartYaw;
                    }
                }
            }
        }
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Called from TurokPlayer::OnPostBeginLevel
	//------------------------------------------------------------------------------------------------------------------------
	void OnMapPostBegin()
	{
        if(g_bGiveStartingWeapons)
        {
            SetStartingWeapon();
        }
        
        int iStartMapIndex = BPR::RandMax(g_iSeedStartMap, g_niStartMapIDs.length());
        int iStartMapID = g_niStartMapIDs[iStartMapIndex];
        int iLevel = GetLevelIndex(iStartMapID);

        switch(Game.GetCurrentMapID())
		{
        case 0:
            BPR::Tasks::Add(TickLevel0);
            break;
        case 5:
            g_bForceLevelPortalActivate = true;
            for(int i = 2; i <= 7; ++i)
            {
                if((g_iSeedHubPortals & (1 << i)) != 0 || iLevel == i)
                {
                    World.TriggerActorsByTID(g_pPlayerActor, i);
                }
            }
            g_bForceLevelPortalActivate = false;
            break;
        case 10:
            World.TriggerActorsByTID(g_pPlayerActor, 3900);
            break;
        case 13:
            World.TriggerActorsByTID(g_pPlayerActor, 666);
            break;
        case 31:
            World.TriggerActorsByTID(g_pPlayerActor, 91);
            break;
        }
        
        bool bMapDataLoaded = LoadMapData(Game.GetCurrentMapID());
        if(g_bStartedGame)
        {
            if(!bMapDataLoaded)
            {
                SetupMapRandomizer();
                SaveMapData(Game.GetCurrentMapID());
            }
            else
            {
                int iCount = 0;
                //find actors and set there positions and sectors
                for(uint i = 0; i < g_npActors.length(); ++i)
                {
                    kActor@ pActor = g_npActors[i];
                    CL_BPR_WorldObject@ pWorldObj = cast<CL_BPR_WorldObject@>(pActor.ScriptObject().obj);
                    for(uint j = 0; j < g_pMapData.actors.length(); ++j)
                    {
                        CL_BPR_MapDataActor@ pMapActor = g_pMapData.actors[j];

                        if(pActor.Type() == pMapActor.iType && pActor.TID() == pMapActor.iTID &&
                            pActor.Origin().DistanceSq(pMapActor.vStartPos) <= 1.0f)
                        {
                            pActor.SetSector(pMapActor.iDestSector);
                            pActor.SetPosition(pMapActor.vDestPos);
                            iCount++;
                            break;
                        }
                    }
                }
            }
        }
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Called from TurokPlayer::OnEndLevel
	//------------------------------------------------------------------------------------------------------------------------
	void OnMapEnd()
	{
        switch(Game.GetCurrentMapID())
		{
        case 42:
            GameVariables.SetValue(GV_KEY_KNIFE_POINTER, "0");
            break;
        }

        BPR::Tasks::OnMapEnd();
		g_npActors.resize(0);
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Called from TurokPlayer::OnTick
	//------------------------------------------------------------------------------------------------------------------------
	void OnMapTick()
	{
        BPR::Input::Tick();
        
        //Keep track of the lowest difficulty the player set the game to
        int iLowestDif;
        if(GameVariables.GetInt(GV_KEY_LOWEST_DIFFICULTY, iLowestDif))
        {
            if(Game.GetDifficulty() < iLowestDif)
            {
                GameVariables.SetValue(GV_KEY_LOWEST_DIFFICULTY, "" + Game.GetDifficulty());
            }
        }
        
		if(!Player.Locked() && !Camera.Active() && BPR::Input::Pressing(BC_MAPZOOMIN|BC_MAPZOOMOUT))
		{
            g_iShowStatsTime = 240;
        }
        
        if(g_iShowStatsTime > 0)
        {
            g_iShowStatsTime--;
            kStr s = "Seed " + BPR::GameVar::GetInt(BPR::GV_KEY_START_SEED) + "\n\n";
            s += BPR::String::FormatHelpTime(BPR::Time::TicksToSeconds());
            Game.PrintHelp(s);
        }
        
        BPR::Tasks::Tick();
        BPR::Time::Tick();
	}
	//------------------------------------------------------------------------------------------------------------------------
	// Called from TurokPlayer::OnPickup
	//------------------------------------------------------------------------------------------------------------------------
	void OnPickup(kActor @pickup)
	{
		
	}
	//------------------------------------------------------------------------------------------------------------------------
    bool PlayerWarpCheck()
    {
		// Warp player to random location when game starts
		if(!Camera.Active())
		{
            SetGameStarted();
            PlayerStartWarp();
            
            return true;
        }
        
        return false;
    }
	//------------------------------------------------------------------------------------------------------------------------
    bool TickLevel0()
    {
        kActor@ pCampaigner = World.GetActorByTID(400);   
        if(pCampaigner is null)
        {
            return false;
        }
        
        if(pCampaigner.Health() <= 0)
        {
            int iInventoryPercent = GetInventoryPercent();
            int iLowestDif = BPR::GameVar::GetInt(BPR::GV_KEY_LOWEST_DIFFICULTY);
            kStr sTime = "Completion Time: ";
            sTime += BPR::String::FormatHMMSSmm(BPR::Time::TicksToSeconds());
            kStr sModTitle = "Randomizer v";
            kStr sLowestDif = "Difficulty: ";
            kStr sDifExtra = "";
            if(Game.GetDifficulty() != iLowestDif)
            {
                sDifExtra += " (You changed difficulty mid game to ";
                sDifExtra += GetDifficultyName(Game.GetDifficulty());
                sDifExtra += ")";
            }

            Sys.Print(" ");
            Sys.Print(sModTitle + VERSION);
			Sys.Print("-------------------------------------------------------");
            Sys.Warning(sTime);
            Sys.Warning("Completion Percent: " + iInventoryPercent);
            Sys.Warning("Seed: " + BPR::GameVar::GetInt(BPR::GV_KEY_START_SEED));
            Sys.Warning(sLowestDif + GetDifficultyName(iLowestDif) + sDifExtra);
			Sys.Print("-------------------------------------------------------");
            Sys.Print(" ");
            
            return true;
        }
        
        return false;
    }
	//------------------------------------------------------------------------------------------------------------------------
    kStr GetDifficultyName(const int dif)
    {
        switch(dif)
        {
        case 0:
            return "Easy";
        case 1:
            return "Normal";
        case 2:
            return "Hard";
        case 3:
            return "Hardcore";
        }
        
        return "Unknown";
    }
	//------------------------------------------------------------------------------------------------------------------------
    // Returns int value from 0 to 100
    int GetInventoryPercent()
    {
        //Get Total number of inventory items
        int iTotalItems = 8; //chrono pieces
        int iItems = 0; //items player currently has
        for(int iHubID = 0; iHubID < 8; ++iHubID)
        {
            int nKeys, remainingKeys, keyBits;
            Game.GetHubKeyInfo(iHubID, nKeys, remainingKeys, keyBits);
            iTotalItems += nKeys;
        }

        //Get keys obtained
        for(int iHubID = 0; iHubID < 8; ++iHubID)
        {
            int nKeys, remainingKeys, keyBits;
            Game.GetHubKeyInfo(iHubID, nKeys, remainingKeys, keyBits);
            iItems += nKeys - remainingKeys;
        }
        
        //Check for chrono pieces
        int chronoBits;
        GameVariables.GetInt("chronoPieceFlags", chronoBits);
        for(int bit = 0; bit < 8; ++bit)
        {
            if((chronoBits & (1 << bit)) != 0)
            {
                iItems++;
            }
        }
        
        if(iItems >= iTotalItems)
        {
            return 100;
        }
        
        return int((iItems / float(iTotalItems)) * 100);
    }
	//------------------------------------------------------------------------------------------------------------------------
    void PlayerStartWarp()
    {
        GameVariables.SetValue(GV_KEY_PWARP, "1");
        bool b = BPR::GameVar::GetBool(GV_KEY_OPT_PLAYER_LOC);
        if(!BPR::GameVar::GetBool(GV_KEY_OPT_PLAYER_LOC))
        {
            Game.ChangeMap("levels/level04.map");
        }
        else
        {
            int iStartMapIndex = BPR::RandMax(g_iSeedStartMap, g_niStartMapIDs.length());
            Game.ChangeMap(g_nsStartMapPaths[iStartMapIndex]);
        }
    }
	//------------------------------------------------------------------------------------------------------------------------
    // Changes certain actor positions with other actors
	//------------------------------------------------------------------------------------------------------------------------
    void SetupMapRandomizer()
    {
        BPR::SetRandSeed(BPR::GameVar::GetInt(GV_KEY_START_SEED) + Game.GetCurrentMapID());
        int iBossMapIndex = g_niBossMapIDs.find(Game.GetCurrentMapID());

        //player: 0, classname "kexPuppet"
		//ambience sounds: 5000-5010, classname "kexAmbience"
		//ammo: 420-436, classname "kexAmmoPickup"
		//animals: 14,20,22,23, classname "kexActor
		//Bosses: 1000-1004, classname "kexActor
		//emitters: 5011-5014,5025-5029,5031-5035,5347-5351,5412,5555, classname "kexEmitter
		//gibs: 701-708, classname "kexActor
		//health: 400-405, classname "kexHealthPickup"
		//humans: 2, classname "kexAI"
		//keys: 440-446,448-449, classname "kexKeyPickup"
		//misc: -1,301-302,305,308-309,317-321,323,325,330,332,600,800,4000, classname "kexActor"
        //monsters: 1,4-5,8-12,15-16,21,25,2001, classname "kexAI"
        //movers:300,303,304,310-311,316,324,328-329, classname "kexActor
        //pickups:406-407,437-439,447,450-456, classname kexLifeForcePickup, kexBackpackPickup, kexSpiritualPickup, kexArmorPickup, kexPickup
        //weapons: 409-419, classname "kexWeaponPickup"
        
        array<kActor@> npValidAIActors;
        for(uint i = 0; i < g_npActors.length(); ++i)
        {
            kActor@ pActor = g_npActors[i];
            if(pActor.SectorIndex() <= -1)
            {
                continue;
            }
            
            kDictMem@ def = pActor.Definition();
            if(def is null)
            {
                continue;
            }
            
            int iTargetTID = pActor.SpawnParams(SP_TARGETID);
            int iTriggerAnim = pActor.SpawnParams(SP_TRIGGERANIM);
            if(iTargetTID > 0 || iTriggerAnim > 0)
            {
                continue;
            }
            
            if((World.GetAreaFlags(pActor.AreaID()) & (AAF_BLOCK|AAF_CLIFF|AAF_CLIMB|AAF_DEATHPIT|AAF_LAVA|AAF_ANTIGRAVITY|AAF_LADDER)) != 0)
            {
                continue;
            }
            
            if(pActor.InWater())
            {
                continue;
            }
            
            kStr sClassName;
            if(!def.GetString("classname", sClassName))
            {
                continue;
            }
            
            if(!BPR::String::Contains(sClassName, "kexAI"))
            {
                continue;
            }
            
            npValidAIActors.insertLast(pActor);
        }
        
        while(npValidAIActors.length() >= 2)
        {
            int iActorIndex = BPR::RandMax(npValidAIActors.length());
            kActor@ pActor = npValidAIActors[iActorIndex];
            npValidAIActors.removeAt(iActorIndex);
            
            iActorIndex = BPR::RandMax(npValidAIActors.length());
            kActor@ pActor2 = npValidAIActors[iActorIndex];
            npValidAIActors.removeAt(iActorIndex);
                        
            kVec3 vActorStartPos = pActor.Origin();
            kVec3 vActor2StartPos = pActor2.Origin();

            int iActorSector = pActor2.SectorIndex();
            kVec3 vActorPos = pActor2.Origin();
            
            int iActor2Sector = pActor.SectorIndex();
            kVec3 vActor2Pos = pActor.Origin();
            
            pActor.SetSector(iActorSector);
            pActor.SetPosition(vActorPos);
            pActor.Flags() &= ~AF_AVOIDWATER;
            pActor2.SetSector(iActor2Sector);
            pActor2.SetPosition(vActor2Pos);
            pActor2.Flags() &= ~AF_AVOIDWATER;
            
            CL_BPR_MapDataActor@ pMapActor1 = CL_BPR_MapDataActor();
            pMapActor1.iType = pActor.Type();
            pMapActor1.iTID = pActor.TID();
            pMapActor1.vStartPos = vActorStartPos;
            pMapActor1.vDestPos = vActorPos;
            pMapActor1.iDestSector = iActorSector;
            g_pMapData.actors.insertLast(pMapActor1);
            
            CL_BPR_MapDataActor@ pMapActor2 = CL_BPR_MapDataActor();
            pMapActor2.iType = pActor2.Type();
            pMapActor2.iTID = pActor2.TID();
            pMapActor2.vStartPos = vActor2StartPos;
            pMapActor2.vDestPos = vActor2Pos;
            pMapActor2.iDestSector = iActor2Sector;
            g_pMapData.actors.insertLast(pMapActor2);
        }

        array<kActor@> npValidActors;
        for(uint i = 0; i < g_npActors.length(); ++i)
        {
            kActor@ pActor = g_npActors[i];
            if(pActor.SectorIndex() <= -1)
            {
                continue;
            }
            
            //Skip actors over death pits
            // int iActorAreaID = pActor.AreaID();
            // uint dwActorAreaFlags = pActor.GetAreaFlags(iActorAreaID);
            // if((dwActorAreaFlags & AAF_DEATHPIT) != 0)
            // {
            // }
            
            kDictMem@ def = pActor.Definition();
            if(def is null)
            {
                continue;
            }
            
            kStr sClassName;
            if(!def.GetString("classname", sClassName))
            {
                continue;
            }
            
            if(BPR::String::Contains(sClassName, "kexKeyPickup") && iBossMapIndex >= 0)
            {
                continue;
            }
            
            if(!BPR::String::Contains(sClassName, "kexKeyPickup") &&
                !BPR::String::Contains(sClassName, "kexWeaponPickup") &&
                !BPR::String::Contains(sClassName, "kexLifeForcePickup") &&
                !BPR::String::Contains(sClassName, "kexBackpackPickup") &&
                !BPR::String::Contains(sClassName, "kexSpiritualPickup") &&
                !BPR::String::Contains(sClassName, "kexArmorPickup") &&
                !BPR::String::Contains(sClassName, "kexPickup") &&
                !BPR::String::Contains(sClassName, "kexHealthPickup") &&
                !BPR::String::Contains(sClassName, "kexAmmoPickup") &&
                pActor.Type() != AT_MINIPORTAL)
            {
                continue;
            }
            
            //Ignore pickup actors that are not really reachable
            if(Game.GetCurrentMapID() == 16 &&
                (pActor.Origin().DistanceSq(kVec3(2355, -2103, -395)) <= GAME_SCALE ||
                pActor.Origin().DistanceSq(kVec3(2351, -1174, -395)) <= GAME_SCALE))
            {
                continue;
            }
            
            npValidActors.insertLast(pActor);
        }

        while(npValidActors.length() >= 2)
        {
            int iActorIndex = BPR::RandMax(npValidActors.length());
            kActor@ pActor = npValidActors[iActorIndex];
            npValidActors.removeAt(iActorIndex);
            
            iActorIndex = BPR::RandMax(npValidActors.length());
            kActor@ pActor2 = npValidActors[iActorIndex];
            npValidActors.removeAt(iActorIndex);
                        
            kVec3 vActorStartPos = pActor.Origin();
            kVec3 vActor2StartPos = pActor2.Origin();

            kDictMem@ cActorDef = pActor.Definition();
            kStr sActorClassName;
            cActorDef.GetString("classname", sActorClassName);
            
            kDictMem@ cActor2Def = pActor2.Definition();
            kStr sActor2ClassName;
            cActor2Def.GetString("classname", sActor2ClassName);
            
            bool bActor1IsKey = BPR::String::Contains(sActorClassName, "kexKeyPickup");
            bool bActor2IsKey = BPR::String::Contains(sActor2ClassName, "kexKeyPickup");
            
            int iActorSector = pActor2.SectorIndex();
            kVec3 vActorPos = pActor2.Origin();
             
            if(bActor2IsKey &&
                (!bActor1IsKey && !BPR::String::Contains(sActorClassName, "kexWeaponPickup")))
            {
                vActorPos.z += 40.96f;
            }
            else if(pActor2.Type() == AT_MINIPORTAL)
            {
                vActorPos.z = BPR::Math::Max(pActor2.FloorHeight(), vActorPos.z - 210.0f);
            }
            
            if(bActor1IsKey && vActorPos.z < pActor2.FloorHeight() - pActor2.Height() + (GAME_SCALE * 2))
            // if(vActorPos.z < pActor2.FloorHeight() - pActor2.Height() + (GAME_SCALE * 2))
            {
                vActorPos.z = pActor2.FloorHeight();
            }
            
            int iActor2Sector = pActor.SectorIndex();
            kVec3 vActor2Pos = pActor.Origin();
            if(BPR::String::Contains(sActorClassName, "kexKeyPickup") &&
                (!bActor2IsKey && !BPR::String::Contains(sActor2ClassName, "kexWeaponPickup")))
            {
                vActor2Pos.z += 40.96f;
            }
            else if(pActor.Type() == AT_MINIPORTAL)
            {
                vActor2Pos.z = BPR::Math::Max(pActor.FloorHeight(), vActor2Pos.z - 210.0f);
            }
            
            if(bActor2IsKey && vActor2Pos.z < pActor.FloorHeight() - pActor.Height() + (GAME_SCALE * 2))
            // if(vActor2Pos.z < pActor.FloorHeight() - pActor.Height() + (GAME_SCALE * 2))
            {
                vActor2Pos.z = pActor.FloorHeight();
            }
            
            // if(bActor1IsKey)
            // {
                // Sys.Print("Key at " + vActorPos.x + " " + vActorPos.y + " " + vActorPos.z);
            // }
            // if(bActor2IsKey)
            // {
                // Sys.Print("Key at " + vActor2Pos.x + " " + vActor2Pos.y + " " + vActor2Pos.z);
            // }

            pActor.SetSector(iActorSector);
            pActor.SetPosition(vActorPos);
            pActor2.SetSector(iActor2Sector);
            pActor2.SetPosition(vActor2Pos);
            
            CL_BPR_MapDataActor@ pMapActor1 = CL_BPR_MapDataActor();
            pMapActor1.iType = pActor.Type();
            pMapActor1.iTID = pActor.TID();
            pMapActor1.vStartPos = vActorStartPos;
            pMapActor1.vDestPos = vActorPos;
            pMapActor1.iDestSector = iActorSector;
            g_pMapData.actors.insertLast(pMapActor1);
            
            CL_BPR_MapDataActor@ pMapActor2 = CL_BPR_MapDataActor();
            pMapActor2.iType = pActor2.Type();
            pMapActor2.iTID = pActor2.TID();
            pMapActor2.vStartPos = vActor2StartPos;
            pMapActor2.vDestPos = vActor2Pos;
            pMapActor2.iDestSector = iActor2Sector;
            g_pMapData.actors.insertLast(pMapActor2);
        }
    }
	//------------------------------------------------------------------------------------------------------------------------
    void SetStartSeed(const int iSeed)
    {
		GameVariables.SetValue(GV_KEY_START_SEED, "" + iSeed);
        GetSeedValues();
    }
	//------------------------------------------------------------------------------------------------------------------------
    void GetSeedValues()
    {
        //Get Random Values
        BPR::SetRandSeed(BPR::GameVar::GetInt(GV_KEY_START_SEED));
        g_iSeedStartMap = BPR::Rand();
        g_iSeedWarp = BPR::Rand();
        g_iSeedHubPortals = BPR::Rand();
        g_iBowShotType = BPR::RandMax(4);
        if(BPR::GameVar::GetBool(GV_KEY_OPT_ALWAYS_BOW))
        {
            g_iBowShotType = 3;
        }
    }
	//------------------------------------------------------------------------------------------------------------------------
    void SetGameStarted()
    {
        g_bStartedGame = true;
        GameVariables.SetValue(GV_KEY_STARTED, "1");
    }
	//------------------------------------------------------------------------------------------------------------------------
    void SetStartingWeapon()
    {
        BPR::SetRandSeed(BPR::GameVar::GetInt(GV_KEY_START_SEED));
        int r = BPR::RandMax(100);
        if(r >= 50)
        {
            //Has Knife and Bow
            Player.GiveWeapon(TW_WEAPON_BOW, 30);
        }
        else if(r >= 20)
        {
            //Has Knife only - Do nothing
            if(BPR::GameVar::GetBool(GV_KEY_OPT_ALWAYS_BOW))
            {
                Player.GiveWeapon(TW_WEAPON_BOW, 30);
            }
        }
        else
        {
            //Has Bow only
            if(!BPR::GameVar::GetBool(GV_KEY_OPT_ALWAYS_KNIFE))
            {
                GameVariables.SetValue(GV_KEY_KNIFE_POINTER, "1");
            }
            Player.GiveWeapon(TW_WEAPON_BOW, 30);
        }
    }
	//------------------------------------------------------------------------------------------------------------------------
    // Returns the Level Number from the MapID
	//------------------------------------------------------------------------------------------------------------------------
    int GetLevelIndex(const int iMapID)
    {
		switch(iMapID)
		{
		case 4:
		case 5:
			return 1;
		case 6:
		case 7:
			return 2;
		case 8:
		case 9:
		case 48:
			return 3;
		case 10:
		case 11:
			return 4;
		case 12:
		case 13:
		case 49:
			return 5;
		case 14:
		case 15:
		case 16:
			return 6;
		case 17:
		case 18:
		case 19:
		case 20:
			return 7;
		case 0:
		case 3:
		case 21:
		case 22:
		case 23:
		case 24:
		case 25:
			return 8;
		}
        
        return 0;
    }
	//------------------------------------------------------------------------------------------------------------------------
    void SetGruntTextures(kActor@ pActor)
    {
        kRenderModel @model = pActor.RenderModel();
        if(model is null)
        {
            return;
        }
        
        if(pActor.ModelVariation() >= GV_CYBORG_PULSE_RIFLE &&
            pActor.ModelVariation() <= GV_CYBORG_SPEAR)
        {
            pActor.ImpactType() = IT_METAL;
        }
        
        switch(pActor.ModelVariation())
        {
        // common soldier
        case GV_GENERIC:
        case GV_GENERIC_AXE:
        case GV_GENERIC_CLUB:
            // face
            model.SetTexture(19, Math::SysRand() % 12);
            break;
        
        // commander
        case GV_COMMANDER_SPEAR:
        case GV_COMMANDER_RIFLE1:
        case GV_COMMANDER_RIFLE2:
        case GV_COMMANDER_RIFLE3:
        case GV_COMMANDER_RIFLE4:
            // chest
            model.SetTexture(0, 1);
            model.SetTexture(9, 1);
            
            // arms
            model.SetTexture(10, 1);
            model.SetTexture(11, 1);
            model.SetTexture(12, 1);
            model.SetTexture(13, 1);
            model.SetTexture(14, 1);
            model.SetTexture(15, 1);
            model.SetTexture(16, 1);
            model.SetTexture(17, 1);
            
            // face
            model.SetTexture(19, Math::SysRand() % 12);
            break;
            
        // poacher
        case GV_POACHER_KNIFE:
        case GV_POACHER_PISTOL1:
        case GV_POACHER_PISTOL2:
        case GV_POACHER_RIFLE:
        case GV_POACHER_SHOTGUN:
            // chest
            model.SetTexture(0, 5);
            model.SetTexture(9, 5);
            
            // legs
            model.SetTexture(1, 2);
            model.SetTexture(3, 2);
            model.SetTexture(4, 2);
            model.SetTexture(5, 2);
            model.SetTexture(7, 2);
            model.SetTexture(8, 2);
            
            // shin
            model.SetTexture(2, 2);
            model.SetTexture(6, 2);
            
            // face
            model.SetTexture(19, Math::SysRand() % 12);
            break;
            
        // warrior
        case GV_WARRIOR_BONE:
        case GV_WARRIOR_DART:
        case GV_WARRIOR_SPEAR:
            // hands
            model.SetTexture(17, 2);
            
            // chest
            model.SetTexture(0, 2);
            model.SetTexture(9, 2);
            
            // arms
            model.SetTexture(10, 2);
            model.SetTexture(11, 2);
            model.SetTexture(12, 2);
            model.SetTexture(13, 2);
            model.SetTexture(14, 2);
            model.SetTexture(15, 2);
            model.SetTexture(16, 2);
            
            // neck
            model.SetTexture(18, 1);
            break;
            
        // shaman
        case GV_SHAMAN:
            // chest
            model.SetTexture(0, 3);
            model.SetTexture(9, 3);
            
            // neck
            model.SetTexture(18, 2);
            
            // legs
            model.SetTexture(1, 1);
            model.SetTexture(3, 1);
            model.SetTexture(4, 1);
            model.SetTexture(5, 1);
            model.SetTexture(7, 1);
            model.SetTexture(8, 1);
            
            // shin
            model.SetTexture(2, 1);
            model.SetTexture(6, 1);
            
            // hands
            model.SetTexture(17, 3);
            
            // arms
            model.SetTexture(10, 3);
            model.SetTexture(11, 3);
            model.SetTexture(12, 3);
            model.SetTexture(13, 3);
            model.SetTexture(14, 3);
            model.SetTexture(15, 3);
            model.SetTexture(16, 3);
            break;
            
        // cyborg guard
        case GV_CYBORG_PULSE_RIFLE:
            // legs
            model.SetTexture(1, 1);
            model.SetTexture(3, 1);
            model.SetTexture(4, 1);
            model.SetTexture(5, 1);
            model.SetTexture(7, 1);
            model.SetTexture(8, 1);
            
            // shin
            model.SetTexture(2, 1);
            model.SetTexture(6, 1);
            
            // chest
            model.SetTexture(0, 4);
            model.SetTexture(9, 4);
            
            // arms
            model.SetTexture(10, 4);
            model.SetTexture(11, 4);
            model.SetTexture(12, 4);
            model.SetTexture(13, 4);
            model.SetTexture(14, 4);
            model.SetTexture(15, 4);
            model.SetTexture(16, 4);
            
            // neck
            model.SetTexture(18, 3);
            break;
            
        // demon
        case GV_DEMON_AXE:
        case GV_DEMON_SPEAR:
        case GV_DEMON_HANDS:
            // hands
            model.SetTexture(17, 6);
            model.SetTexture(13, 6);
            
            // legs
            model.SetTexture(1, 2);
            model.SetTexture(3, 2);
            model.SetTexture(4, 2);
            model.SetTexture(5, 2);
            model.SetTexture(7, 2);
            model.SetTexture(8, 2);
            
            // shin
            model.SetTexture(2, 2);
            model.SetTexture(6, 2);
            
            // chest
            model.SetTexture(0, 6);
            model.SetTexture(9, 6);
            
            // arms
            model.SetTexture(10, 6);
            model.SetTexture(11, 6);
            model.SetTexture(12, 6);
            model.SetTexture(13, 6);
            model.SetTexture(14, 6);
            model.SetTexture(15, 6);
            model.SetTexture(16, 6);
            
            // neck
            model.SetTexture(18, 4);
            break;
            
        // demon lord
        case GV_DEMON_LORD:
            // hands
            model.SetTexture(17, 7);
            model.SetTexture(13, 7);
            
            // legs
            model.SetTexture(1, 3);
            model.SetTexture(3, 3);
            model.SetTexture(4, 3);
            model.SetTexture(5, 3);
            model.SetTexture(7, 3);
            model.SetTexture(8, 3);
            
            // shin
            model.SetTexture(2, 3);
            model.SetTexture(6, 3);
            
            // chest
            model.SetTexture(0, 7);
            model.SetTexture(9, 7);
            
            // arms
            model.SetTexture(10, 7);
            model.SetTexture(11, 7);
            model.SetTexture(12, 7);
            model.SetTexture(13, 7);
            model.SetTexture(14, 7);
            model.SetTexture(15, 7);
            model.SetTexture(16, 7);
            
            // neck
            model.SetTexture(18, 5);
            break;
        }
    }
	//------------------------------------------------------------------------------------------------------------------------
    int GetSpawnActorType(kActor@ pActor)
    {        
        kDictMem@ def = pActor.Definition();
        if(def is null)
        {
            return pActor.Type();
        }
        
        int iSpawnActorType;
        if(def.GetInt("BPR_SpawnActorType", iSpawnActorType))
        {
            return iSpawnActorType;
        }
        
        return pActor.Type();
    }
	//------------------------------------------------------------------------------------------------------------------------
    int GetKnifeActorType(kActor@ pActor)
    {
        int actorType = pActor.Type();
        if(actorType != -1)
            return actorType;
        
        kDictMem@ def = pActor.Definition();
        if(def is null)
        {
            return actorType;
        }
        
        int iSpawnActorType;
        if(def.GetInt("BPR_SpawnActorType", iSpawnActorType))
        {
            return iSpawnActorType;
        }
        
        return actorType;
    }
	//------------------------------------------------------------------------------------------------------------------------
    kActor@ SpawnActor(const kStr &in actorName, const float x, const float y, const float z, const float yaw, const int sector = -1)
    {
        BPR::g_bSpawningActor = true;
        kActor@ pActor = ActorFactory.Spawn(actorName, x, y, z, yaw, sector);
        BPR::g_bSpawningActor = false;
        return pActor;
    }
	//------------------------------------------------------------------------------------------------------------------------
}
