namespace BPR { namespace Tasks
{
    // Task will be removed if function returns true
    funcdef bool FUNCDEF_TASK();
    
    array<CL_BPR_Task@> g_pTasks;
    //========================================================================================================================	
    class CL_BPR_Task
    {
        FUNCDEF_TASK@ m_pFunc;
        uint m_iDelay;
        uint m_iRepeatDelay;
        CL_BPR_Task(FUNCDEF_TASK@ pFunc, const uint iStartDelay, const uint iRepeatDelay)
        {
            @m_pFunc = pFunc;
            m_iDelay = iStartDelay;
            m_iRepeatDelay = iRepeatDelay;
        }
    }
    //------------------------------------------------------------------------------------------------------------------------
    // Adds task to the task list
    //------------------------------------------------------------------------------------------------------------------------
    void Add(FUNCDEF_TASK@ pFunc, const uint iStartDelay = 0, const uint iRepeatDelay = 0)
    {
        Assert(pFunc !is null, "[BPR::AddTask] Cannot add a task with a null function");
        g_pTasks.insertLast(CL_BPR_Task(pFunc, iStartDelay, iRepeatDelay));
    }
    //------------------------------------------------------------------------------------------------------------------------
    // Adds task if pFunc isn't already in the task list. Returns true if added to the task list.
    //------------------------------------------------------------------------------------------------------------------------
    bool AddUnique(FUNCDEF_TASK@ pFunc, const uint iStartDelay = 0, const uint iRepeatDelay = 0)
    {
        for(int i = int(g_pTasks.length()) - 1; i >= 0; i--)
        {
            if(@g_pTasks[i].m_pFunc == @pFunc)
            {
                return false;
            }
        }

        Add(pFunc, iStartDelay, iRepeatDelay);
        return true;
    }
    //------------------------------------------------------------------------------------------------------------------------
    void Tick()
    {
        for(int i = int(g_pTasks.length()) - 1; i >= 0; i--)
        {
            CL_BPR_Task@ pTask = g_pTasks[i];
            if(pTask.m_iDelay == 0)
            {
                if(pTask.m_pFunc())
                {
                    g_pTasks.removeAt(i);
                }
                else
                {
                    pTask.m_iDelay = pTask.m_iRepeatDelay;
                }
            }
            else
            {
                pTask.m_iDelay--;
            }
        }
    }
    //------------------------------------------------------------------------------------------------------------------------
    void OnMapEnd()
    {
        g_pTasks.resize(0);
    }
    //------------------------------------------------------------------------------------------------------------------------
}}
